#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This le is part of pspman.
#
# pspman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pspman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pspman.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Personal Simple Python-based package manager that clones git
'''


from os import getcwd, chdir, makedirs
from os.path import isdir, isfile
from shutil import rmtree
from sys import exit as sysexit
from subprocess import Popen, PIPE
from pathlib import Path
from re import compile as recompile
from .classes import InstallEnv
from .input_output import print_info, cli


def git_pulls(env: InstallEnv) -> list:
    '''
    Pull git projects
    '''
    fails = 0
    pull_paths = []
    for git_path in env.git_project_paths:
        chdir(git_path)
        call = Popen(["git", "pull"],
                     stderr=PIPE, stdout=PIPE, text=True)
        stdout, stderr = call.communicate()
        if "Already up to date" not in stdout:
            print_info()
            print_info(f"Updating {git_path}", 1)
            if any(m in stdout for m in ("+", "-")):
                pull_paths.append(git_path)
            else:
                fails += 1
                print_info(f"Failed in {git_path}", 3)
    chdir(env.base_dir)
    for update in pull_paths:
        print_info(f"Updated {update}", 1)
    print_info("", 0)
    if fails:
        print_info(f"{fails} project updates failed", 1)
    return pull_paths


def specific_unknown(_) -> int:
    '''
    Unknown protocol for installation
    '''
    return 0


def specific_make(env: InstallEnv) -> int:
    '''
    install specific make
    return: failure: 1; success: 0
    '''
    if Path("./configure").exists():
        call = Popen(["./configure", "--prefix", env.prefix],
                     stdout=PIPE, stderr=PIPE, text=True)
        stdout, stderr = call.communicate()
        if stderr:
            return 1
    if Path("./Makefile").exists():
        call = Popen(["make"],
                     stdout=PIPE, stderr=PIPE, text=True)
        stdout, stderr = call.communicate()
        if stderr:
            return 1
        call = Popen(["make", "install"],
                     stdout=PIPE, stderr=PIPE)
        stdout, stderr = call.communicate()
        if stderr:
            return 1
    return 0


def specific_pip(_) -> int:
    '''
    specific python setup
    return: failure: 1; success: 0
    '''
    call = Popen(["python", "-m", "pip", "install", "--user", "-U", "."],
                 stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = call.communicate()
    if stderr:
        return 1
    return 0


def specific_meson(env: InstallEnv) -> int:
    ''' specific meson/ninja return: failure: 1; success: 0
    '''
    update_dir = "./build/update"
    makedirs(f'{getcwd}/subprojects', exist_ok=True)
    call = Popen(["gitman", "-c", f"{getcwd()}/subprojects"],
                 stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = call.communicate()
    makedirs(update_dir, exist_ok=True)
    call = Popen(["meson", "--wipe", "--buildtype=release", "--prefix",
                 env.prefix, "-Db_lto=true", update_dir],
                 stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = call.communicate()
    if stderr:
        call = Popen(["meson", "--buildtype=release", "--prefix",
                     env.prefix, "-Db_lto=true", update_dir],
                     stdout=PIPE, stderr=PIPE, text=True)
        stdout, stderr = call.communicate()
        if stderr:
            return 1
    chdir(update_dir)
    call = Popen(["ninja"], stderr=PIPE, stdout=PIPE, text=True)
    stdout, stderr = call.communicate()
    if stderr:
        return 1
    call = Popen(["ninja", "install"], stderr=PIPE, stdout=PIPE, text=True)
    stdout, stderr = call.communicate()
    if stderr:
        return 1
    return 0


def install_wrap(env: InstallEnv, projects_list: list,
                 install_type: int) -> None:
    '''
    install_type: integer indicating installs:
    0: unknown
    1: make
    2: pip
    3: meson
    projects_list: list of projects to be installed

    parent definition object for all kinds of install
    '''
    inst_protocol = [
        "Unknown",
        "./configure -> make -> make install",
        "pip install --user -U .",
        "meson/ninja",
    ]
    print_info("", 0)
    print_info("Looks like the method of installation is", 1)
    print_info(inst_protocol[install_type], 0)
    print_info("for the following projects", 0)
    print_info("", 0)
    for proj in projects_list:
        print_info(proj, 5)
        chdir(proj)
        # TODO: UPDATE Flag if already present
        print_info(f"cd {getcwd()}", 1)
        err = [specific_unknown, specific_make,
               specific_pip, specific_meson][install_type](env)
        if err:
            print_info("Failed", 4)
            continue
    chdir(env.base_dir)
    print_info("", 0)


def auto_install(git_paths: list, env: InstallEnv) -> None:
    unknown_installs = []
    make_projects = []
    meson_projects = []
    pip_projects = []
    protocols = (
        unknown_installs,
        make_projects,
        pip_projects,
        meson_projects,
    )
    for git_path in git_paths:
        if git_path.joinpath("Makefile").exists():
            make_projects.append(git_path)
        elif git_path.joinpath('configure').exists():
            make_projects.append(git_path)
        elif git_path.joinpath('setup.py').exists():
            pip_projects.append(git_path)
        elif git_path.joinpath('meson.build').exists():
            meson_projects.append(git_path)
        else:
            unknown_installs.append(git_path)
    for index, proj_protocol in enumerate(reversed(protocols)):
        if proj_protocol:
            install_wrap(env=env, projects_list=proj_protocol,
                         install_type=len(protocols) - index - 1)


def new_install(env: InstallEnv) -> None:
    makedirs(env.clonedir, exist_ok=True)
    node_pat = recompile("(.*?)/")
    clone_paths_list = []
    for url in env.pkg_install:
        if url[-1] != "/":
            url += "/"
        package = node_pat.findall(url)[-1].replace(".git", "").split(":")[-1]
        package_dir = Path.joinpath(env.clonedir, package)
        if isdir(package_dir):
            print_info(f"{package} appears to be installed already", 3)
        elif isfile(package_dir):
            print_info(f"A file named '{package_dir}' already exists", 3)
            package_dir = package_dir.joinpath(".d")
            if package_dir.exists():
                print_info("", 0)
                print_info(f"{package_dir} also exists,", 3)
                print_info("This is too much to handle...", 4)
                continue
            print_info(f"Calling this project '{package_dir}'", 3)
            print_info(f"Installing in {package_dir}", 1)
            chdir(env.clonedir)
            makedirs(package_dir, exist_ok=False)
            call = Popen(["git", "clone", url, str(package_dir)],
                         stdout=PIPE, stderr=PIPE, text=True)
            stdout, stderr = call.communicate()
            clone_paths_list.append(package_dir)
            if stderr:
                return False
    auto_install(clone_paths_list, env)


def del_proj(env: InstallEnv) -> None:
    for package in env.pkg_delete:
        pkg_path = Path.joinpath(env.clonedir, package)
        if not isdir(pkg_path):
            print_info(f"Couldn't find {package} in {env.clonedir}", 3)
            print_info("Ignoring...", 0)
            continue
        chdir(pkg_path)
        call = Popen(["git", "remote", "-v"],
                     stdout=PIPE, stderr=PIPE, text=True)
        stdout, stderr = call.communicate()
        if stderr:
            return False
        fetch_source = recompile(r"^.*fetch.*").findall(
            stdout)[0].split(" ")[-2].split("\t")[-1]
        print_info(f"Deleting {pkg_path}", 1)
        print_info("I can't guess which files were installed.", 1)
        print_info("So, leaving those scars behind...", 0)
        print_info("This project may be added again from the following path",
                   0)
        print_info(f"{fetch_source}", 0)
        chdir(env.clonedir)
        rmtree(Path.joinpath(env.clonedir, package))


def main() -> None:
    args = cli()
    pkg_grp = InstallEnv(
        clonedir=args.clone_dir,
        prefix=args.prefix,
        only_pull=args.only_pull,
        force_root=args.force_root,
        pkg_install=args.pkg_install or "",
        pkg_delete=args.pkg_delete or "",
    )
    del_proj(env=pkg_grp)
    pkg_grp.find_gits()
    updates = git_pulls(env=pkg_grp)
    auto_install(updates, pkg_grp)
    new_install(pkg_grp)
    chdir(pkg_grp.base_dir)
    print_info("done.", 1)
    sysexit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This le is part of pspman.
#
# pspman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pspman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pspman.  If not, see <https://www.gnu.org/licenses/>.
#
'''
input/outputs
'''


from os import environ
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, Namespace
from colorama import Fore, Style


def print_info(out_msg: str = '', p_type: int = 0, **kwargs) -> None:
    '''
    out_msg: Printed with an prefix of p_type
    p_type: interpreted type to prefix out_msg
    0: continuation
    1: information
    2: action
    3: warning
    4: error
    5: list
    6: debug
    *: <empty>
    **kwargs: passed to print function
    '''
    emit_l = [
        Style.RESET_ALL + '          ' + Style.RESET_ALL,
        Fore.GREEN +      ' [INFORM] ' + Style.RESET_ALL,
        Fore.YELLOW +     ' [ACTION] ' + Style.RESET_ALL,
        Fore.MAGENTA +    '[WARNING] ' + Style.RESET_ALL,
        Fore.RED +        '  [ERROR] ' + Style.RESET_ALL,
        Fore.BLUE +       '   [LIST] ' + Style.RESET_ALL,
        Fore.CYAN +       '  [DEBUG] ' + Style.RESET_ALL,
    ]
    if not (0 <= p_type <= 6):
        p_type = 0
    print(emit_l[p_type] + str(out_msg), **kwargs, flush=True)


def cli() -> Namespace:
    '''
    Parse command line arguments
    '''
    description = '''
    NOTICE: This is only intended for \"user\" installs.
    CAUTION: DO NOT RUN THIS SCRIPT AS ROOT.
    CAUTION: If you still insist, I won't care.
    '''
    homedir = environ["HOME"]
    parser = ArgumentParser(description=description,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-f', '--force-root', action='store_true',
                        help='force working with root permissions [DANGEROUS]')
    parser.add_argument('-o', '--only-pull', action='store_true',
                        help='only pull, do not try to install')
    parser.add_argument('-c', '--clone-dir', type=str, nargs="?",
                        default=f'{homedir}/.pspman/programs',
                        help='path for all git clones')
    parser.add_argument('-p', '--prefix', type=str, nargs="?",
                        default=f'{homedir}/.pspman',
                        help='path for installation')
    parser.add_argument('-i', '--pkg-install', metavar='URL', type=str,
                        nargs="*", help='URL to clone new project')
    parser.add_argument('-d', '--pkg-delete', metavar='PROJ', type=str,
                        nargs="*", help='PROJ to clone new project')
    args = parser.parse_args()
    return args



#!/usr/bin/env python3
# -*- coding:utf-8; mode:python -*-
#
# Copyright 2020 Pradyumna Paranjape
# This le is part of pspman.
#
# pspman is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pspman is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pspman.  If not, see <https://www.gnu.org/licenses/>.
#
'''
Environment Classes
'''


from os import environ, listdir, getcwd
from os.path import isdir
from sys import exit as sysexit
from time import sleep
from subprocess import Popen, PIPE
from pathlib import Path
from .input_output import print_info


def timeout(wait: int = 10) -> None:
    '''
    print a count-down time that waits for the process to get killed
    '''
    for time_out in range(wait):
        sleep(1)
        print(f"{wait - time_out:2.0f} seconds...", end="", flush=True)
        print("\b" * 13, end="", flush=True)
    print(f"{wait:2.0f} seconds...", flush=True)


class InstallEnv():
    '''
    Installation Environment
    '''
    def __init__(self, clonedir: str, prefix: str, **kwargs):
        '''
        clonedir: Directory in which gits are cloned and maintained
        prefix: Directory in which "bin", "share", "lib" are stored
        force_root: Force installation as ROOT (Administrator)
        only_pull: Don't try to install
        pkg_install: List of new packages to install
        pkg_delete: List of new packages to delete

        '''
        self.clonedir = Path(clonedir)
        self.prefix = Path(prefix)
        self.only_pull = kwargs["only_pull"]
        self.force_root = kwargs["force_root"]
        self.pkg_install = list(kwargs.get("pkg_install", ""))
        self.pkg_delete = list(kwargs.get("pkg_delete", ""))
        self.base_dir = getcwd()
        self.git_project_paths = []
        self.permission_check()

    def find_gits(self) -> list:
        '''
        locate git projects
        '''
        count = 0
        for leaf in listdir(self.clonedir):
            if isdir(Path.joinpath(self.clonedir, leaf)):
                if isdir(Path.joinpath(self.clonedir, leaf, ".git")):
                    self.git_project_paths.append(
                        Path.joinpath(self.clonedir, leaf)
                    )
                    count += 1
        print_info(f"{count} project(s) found in {str(self.clonedir)}", 1)

    def permission_check(self) -> None:
        '''
        check permissions in the given context
        '''
        # Am I root?
        if environ["USER"].lower() == "root":
            print_info("I hate dictators", 3)
            if not self.force_root:
                print_info("Bye", 0)
                sysexit(2)
            print_info("I can only hope you know what you are doing...", 3)
            print_info("Here is a chance to kill me in", 2)
            timeout(10)
            print_info("", 0)
            print_info("¯\_(ツ)_/¯ Your decision ¯\_(ツ)_/¯", 3)
            print_info("", 0)
            print_info("[INFO] Proceeding...", 1)
        else:
            # Is installation directory read/writable
            parentdir = Path(self.clonedir)
            while not isdir(parentdir):
                parentdir = Path(parentdir).parent
            self.perm_pass(parentdir)
            self.perm_pass(self.prefix)

    @staticmethod
    def perm_pass(permdir: str) -> None:
        '''
        permdir: directory whose permissions are to be checked
        returns: True if all rwx permissions are granted
        '''
        call = Popen(["stat", "-L", "-c", "%U %G %a", permdir],
                     stderr=PIPE, stdout=PIPE, text=True)
        stdout, stderr = call.communicate()
        if stderr:
            print_info(f"{stderr}", 4)
            print_info("aborting...", 4)
            sysexit()
        owner, group, octperm = stdout.replace("\n", "").split(" ")
        if (octperm[-1] == "7") != 0:
            # everyone has access
            return True
        if (octperm[-2] == "7") != 0:
            # some group has permissions
            call = Popen(["groups", environ["USER"]],
                         stderr=PIPE, stdout=PIPE, text=True)
            stdout, stderr = call.communicate()
            user_groups = stdout.split(" ")
            for ug in user_groups:
                if ug == group:
                    return True
        if (octperm[-3] == "7") != 0:
            # owner has permissions
            if environ["USER"] == owner:
                return True
        print_info("We do not have sufficient permissions", 4)
        print_info("Try another location", 2)
        print_info("Bye", 0)
        sysexit(1)
        return False
